@@ -2,7 +2,7 @@
 ## (C)2006 Simon Urbanek <simon.urbanek@r-project.org>
 ## For license terms see DESCRIPTION and/or LICENSE
 ##
-## $Id: jinit.R 303 2007-06-23 06:51:12Z urbanek $
+## $Id: jinit.R 319 2008-09-20 20:21:06Z urbanek $
 
 .check.JVM <- function() 
     .Call("RJava_checkJVM", PACKAGE="rJava")
@@ -55,7 +55,7 @@
 
   # this should remove any lingering .jclass objects from the global env
   # left there by previous versions of rJava
-  pj <- grep("^\\.jclass",ls(1,all=TRUE),value=TRUE)
+  pj <- grep("^\\.jclass",ls(1,all.names=TRUE),value=TRUE)
   if (length(pj)>0) { 
     rm(list=pj,pos=1)
     if (exists(".jniInitialized",1)) rm(list=".jniInitialized",pos=1)
